             .title        "SxROM Reader Plugin"


;06/15/00 
;Written by KH
;Version 1.0
             
;07/20/06
;BootGod
;Added support for user vars to override PRG / CHR size detection. Code does not check for "valid" values, up to user to do it right.
;Sizes are in terms of 256 byte pages.
;lo | hi | size
;--------------
;20 | 00 |   8K
;40 | 00 |  16K
;80 | 00 |  32K
;00 | 01 |  64K
;00 | 02 | 128K
;00 | 04 | 256K
;00 | 08 | 512K

;01/23/09
;Kinopio
;Cleaned up the code and added support for SXROM and SUROM.
;This plugin should now handle all known S*ROM carts except for SFEXPROM.

;05/10/09
;BootGod
;Added support for 8K CHR. Seta thought they'd be funny and stick a CHR ROM in a SNROM board :/
;Was dumpable via user vars for CHR override, but auto-detection is always better!


             ;vectors for standard system calls

send_byte:   .equ 0200h
baton:       .equ 0203h
chk_vram:    .equ 0206h
chk_wram:    .equ 0209h
wr_ppu:      .equ 020ch
read_byte:   .equ 020fh
init_crc:    .equ 0212h
do_crc:      .equ 0215h
finish_crc:  .equ 0218h

crc0:        .equ 0080h
crc1:        .equ 0081h
crc2:        .equ 0082h
crc3:        .equ 0083h

readaddr:    .equ 00e0h
readaddr_lo: .equ 00e0h
readaddr_hi: .equ 00e1h

banks_left:  .equ 00e2h
cur_bank:    .equ 00e3h

wram0:       .equ 00e4h
wram1:       .equ 00e5h

tempvar:     .equ 00e6h
temp_crc:    .equ 00e7h

uv1enabled:   .equ 07ceh
poverride_lo: .equ 07cfh
uv2enabled:   .equ 07deh
poverride_hi: .equ 07dfh
uv3enabled:   .equ 07eeh
coverride_lo: .equ 07efh
uv4enabled:   .equ 07feh
coverride_hi: .equ 07ffh

;plugin header defines
p_ver: = 4     ;plugin version
p_read: = 1    ;plugin is for downloading (dumping)
p_write: = 2   ;plugin is for uploading (RAM carts / WRAM etc)
p_other: = 3   ;plugin is for some other purpose (i.e. BankWatch)
p_var1: = 4    ;plugin uses user variable #1
p_var2: = 8    ;plugin uses user variable #2
p_var3: = 16   ;plugin uses user variable #3
p_var4: = 32   ;plugin uses user variable #4

;plugin header data

             .org 0380h
             
             ;plugin description
             .db "Nintendo SxROM / MMC1 carts"
             .fill 03e0h-*,00h

             ;plugin creator(s)
             .db "Kevtris,BootGod,Kinopio"
             .fill 03f8h-*,00h

             ;PRG/CHR/WRAM min/max sizes supported
             .db 32 >> 4,512 >> 4
             .db 8 >> 3,128 >> 3
             .db 8,32

             ;plugin version and flags
             .db p_ver
             .db p_read | p_var1 | p_var2 | p_var3 | p_var4

;plugin code start here

             lda #04h          ;mapper controlled mirroring
             jsr send_byte     ;send mirroring info
             
             lda #80h
             sta 0ffffh        ;reset MMC1
             lda #00h
             jsr wr_mmc0       ;8K CHR, 32K PRG (PRG/CHR bank = 16/4K)
             jsr wr_mmc1
             
;detect PRG size

             lda uv1enabled
             cmp #0ffh
             bne prg_override

             lda #00h
             jsr do_prg_crc    ;CRC PRG bank 0
             jsr store_crc     ;store CRC of first 32K PRG (bank 0+1)
             
             lda #10h          ;detect from max to min size, 512->32K
             sta banks_left    ;512K PRG. 1 bank = 32K
             
             jsr chk_vram      ;if CHR is RAM check if 512K PRG (for SUROM and SXROM boards)
             bne detp_loop1    ;CHR isn't RAM, keep checking

             lda #10h
             jsr wr_mmc1       ;try and select next 256K bank
             lda #00h
             jsr do_prg_crc    ;CRC PRG bank 0 (16)
             jsr comp_crc      ;check if CRC of PRG bank 0 changed
             bne got_prg
                          
detp_loop1:  lsr banks_left    ;halve size
             lda banks_left
             cmp #01h
             beq got_prg       ;32K
             jsr do_prg_crc    ;CRC PRG bank; 08h, 04h, 02h
             jsr comp_crc      ;compare CRC with PRG bank 0
             bne got_prg       ;if not equal then PRG size found; 256K, 128K, 64K
             jmp detp_loop1    ;keep looping until size is found
             
;dump PRG

prg_override:lda poverride_lo
             jsr send_byte
             lda poverride_hi
             jsr send_byte

             asl poverride_hi
             lda poverride_lo
             rol a
             lda #0            ;???
             rol a
             ora poverride_hi
             sta banks_left    ;was tay in old code
             
             jmp blktypep

got_prg:     lda banks_left
             sta tempvar
             lda #00h
             lsr tempvar
             ror a
             jsr send_byte
             lda tempvar
             jsr send_byte     ;send size
blktypep:    lda #01h
             jsr send_byte     ;type is PRG
             
             lda #00h
             sta cur_bank      ;init bank variables
             jsr wr_mmc1
             

dmpp_loop1:  lda #00h
             sta readaddr_lo
             tay
             lda #80h
             sta readaddr_hi   ;readaddr=$8000h
             
             lda cur_bank
             cmp #10h          ;time to change 256K bank? Bank 0-15 first 256K, 16-31 last 256K
             bne no_switch
             lda #10h
             jsr wr_mmc1
no_switch:   lda cur_bank
             jsr wr_mmc3       ;select bank 32K at $8000h

dmpp_loop2:  lda (readaddr),y
             jsr send_byte
             iny
             bne dmpp_loop2    ;send a byte 256 times
             jsr baton
             inc readaddr_hi
             bne dmpp_loop2    ;repeat 128 times, $8000h->$FFFFh, 128*256=32768
             inc cur_bank
             inc cur_bank      ;2*16K
             dec banks_left
             bne dmpp_loop1
             
;detect CHR size

             jsr chk_vram      ;check if CHR is RAM
             bne no_vram
             jmp no_chr_rom    ;nothing to dump then
             
no_vram:     lda uv3enabled
             cmp #0ffh
             bne chr_override
             
             lda #00h
             jsr do_chr_crc    ;CRC CHR bank 0
             jsr store_crc     ;store CRC of first 8K CHR (bank 0+1)
             
             lda #10h          ;detect from max to min size, 128->16K
             sta banks_left    ;128K CHR. 1 bank = 8K
             
detc_loop1:  lda banks_left
             cmp #01h          ;Changed from 2 to 1 (16K -> 8K) -BootGod
             beq got_chr       ;8K
             jsr do_chr_crc    ;CRC CHR bank; 10h, 08h, 04h
             jsr comp_crc      ;compare CRC with CHR bank 0
             bne got_chr       ;if not equal then CHR size found; 128K, 64K, 32K
             lsr banks_left    ;halve size
             jmp detc_loop1    ;keep looping until size is found
             
;dump CHR

chr_override:lda coverride_lo
             jsr send_byte
             lda coverride_hi
             jsr send_byte

             asl coverride_hi
             asl coverride_hi
             asl coverride_hi
             lda coverride_lo
             lsr a
             lsr a
             lsr a
             lsr a
             lsr a
             ora coverride_hi
             sta banks_left    ;was tay in old code

             jmp blktypec

got_chr:     lda banks_left
             sta tempvar
             lda #00h
             lsr tempvar
             ror a
             lsr tempvar
             ror a
             lsr tempvar
             ror a
             jsr send_byte
             lda tempvar
             jsr send_byte     ;send size
blktypec:    lda #02h
             jsr send_byte     ;type is CHR
             
             lda #00h
             sta cur_bank      ;init bank variables
             
dmpc_loop1:  lda cur_bank
             jsr wr_mmc1
             ldx #20h
             lda #00h
             tay
             sta 2006h
             sta 2006h
             lda 2007h

dmpc_loop2:  lda 2007h
             jsr send_byte
             iny
             bne dmpc_loop2    ;send a byte 256 times
             jsr baton
             dex
             bne dmpc_loop2    ;repeat 32 times, $0000h->$1FFFh, 32*256=8192
             inc cur_bank
             inc cur_bank      ;2*4K
             dec banks_left
             bne dmpc_loop1
             
;detect WRAM size

no_chr_rom:  lda #00h
             jsr wr_mmc3       ;turn RAM enable on (note: MMC1B/C only!!)
             lda #00h
             jsr wr_mmc1       ;select first WRAM bank (SUROM, SOROM)
             jsr chk_wram
             beq yes_wram
             jmp done          ;if no wram then nothing more to do

yes_wram:    lda 6000h
             sta wram0         ;save WRAM content
             lda #04h          ;32K
             sta 6000h         ;store number of 8K banks to dump.
			                   ;SOROM has 16K but only 8K is battery backed.
             
             lda #04h
             jsr wr_mmc1       ;select second WRAM bank (SXROM)
             lda 6000h
             sta wram1         ;save WRAM content, same as wram0 if 8K or 16K.
             lda #01h          ;8K
             sta 6000h         ;if 8K or 16K WRAM then previous value will be overwritten
             
                               ;restore WRAM content and get number of banks
             lda #00h
             jsr wr_mmc1
             lda 6000h
             sta tempvar       ;6000h should now contain the number of 8K banks
             sta banks_left
             lda wram0
             sta 6000h
             
             lda #04h
             jsr wr_mmc1
             lda wram1
             sta 6000h

;dump WRAM
             
             lda #00h
             lsr tempvar
             ror a
             lsr tempvar
             ror a
             lsr tempvar
             ror a
             jsr send_byte
             lda tempvar       ;tempvar should be 00h now
             jsr send_byte     ;send size
             lda #03h
             jsr send_byte     ;type is WRAM
             
			 lda #00h
			 sta cur_bank
             lda banks_left
             cmp #01h          ;check if 8K
             bne dmpw_loop1
             lda #08h          ;start with upper 8K for SOROM, doesn't affect 8K carts. 
			 sta cur_bank

dmpw_loop1:  lda cur_bank			 
             jsr wr_mmc1

             ldx #20h
             lda #00h
             sta readaddr_lo
             tay
             lda #60h
             sta readaddr_hi   ;readaddr=$6000h

dmpw_loop2:  lda (readaddr),y
             jsr send_byte
             iny
             bne dmpw_loop2    ;send byte 256 times
             jsr baton
             inc readaddr_hi
             dex
             bne dmpw_loop2    ;repeat 32 times, $6000h->$7FFFh, 32*256=8192
             
             inc cur_bank
             inc cur_bank
             inc cur_bank
			 inc cur_bank

             dec banks_left
             bne dmpw_loop1

;we are done

done:        lda #00h
             jsr send_byte
             lda #00h
             jsr send_byte
             lda #00h
             jsr send_byte     ;send we are done

             rts

 
;calculate CRC of 32K PRG bank specified by accumulator

do_prg_crc:  jsr wr_mmc3
             lda #00h
             tay
             sta readaddr_lo
             lda #80h
             sta readaddr_hi   ;readaddr=$8000h
             jsr init_crc

crcp_loop:   lda (readaddr),y
             jsr do_crc
             iny
             bne crcp_loop     ;repeat 256 times
             inc readaddr_hi
             bne crcp_loop     ;repeat 128 times, 128*256=32768
             jsr finish_crc
             
             rts
             
;calculate CRC of 8K CHR bank specified by accumulator

do_chr_crc:  jsr wr_mmc1
             lda #00h
             tay
             sta 2006h
             sta 2006h
             lda #20h
             sta tempvar
             jsr init_crc
             lda 2007h

crcc_loop:   lda 2007h
             jsr do_crc
             iny
             bne crcc_loop     ;repeat 256 times
             dec tempvar
             bne crcc_loop     ;repeat 32 times, 32*256=8192
             jsr finish_crc

             rts

;store current CRC in temp_crc

store_crc:   ldx #3

st_loop:     lda crc0,x
             sta temp_crc,x
             dex
             bpl st_loop
             
             rts

;compare current CRC with temp_crc, if they match return 0 else return !0
    
comp_crc:    ldx #3

co_loop:     lda crc0,x
             cmp temp_crc,x
             bne no_comp
             dex
             bpl co_loop
             lda #0

no_comp:     rts

;MMC1 write register functions

wr_mmc0:     sta 09fffh             
             lsr a
             sta 09fffh
             lsr a
             sta 09fffh
             lsr a
             sta 09fffh
             lsr a
             sta 09fffh
             rts
             
wr_mmc1:     pha
             jsr wr_mmc2
             pla
             sta 0bfffh
             lsr a
             sta 0bfffh
             lsr a
             sta 0bfffh
             lsr a
             sta 0bfffh
             lsr a
             sta 0bfffh
             rts
             
wr_mmc2:     sta 0dfffh
             lsr a
             sta 0dfffh
             lsr a
             sta 0dfffh
             lsr a
             sta 0dfffh
             lsr a
             sta 0dfffh
             rts
             
wr_mmc3:     sta 0ffffh             
             lsr a
             sta 0ffffh
             lsr a
             sta 0ffffh
             lsr a
             sta 0ffffh
             lsr a
             sta 0ffffh
             rts
             
             .fill 07c0h-*,0ffh   ;fill rest to get 1K of data

             .db "PRG Pages (lo)"
             .fill 07ceh-*,0
             .fill 07d0h-*,0ffh

             .db "PRG Pages (hi)"
             .fill 07deh-*,0
             .fill 07e0h-*,0ffh

             .db "CHR Pages (lo)"
             .fill 07eeh-*,0
             .fill 07f0h-*,0ffh

             .db "CHR Pages (hi)"
             .fill 07feh-*,0
             .fill 0800h-*,0ffh
             
             .end
             